import numpy as np
import polars as pl
import pytest
import typeguard as tg

from mesa_frames import GridPolars, ModelDF
from tests.pandas.test_agentset_pandas import (
    ExampleAgentSetPandas,
    fix1_AgentSetPandas,
)
from tests.polars.test_agentset_polars import (
    ExampleAgentSetPolars,
    fix2_AgentSetPolars,
)


# This serves otherwise ruff complains about the two fixtures not being used
def not_called():
    fix1_AgentSetPandas()
    fix2_AgentSetPolars()


@tg.typechecked
class TestGridPolars:
    @pytest.fixture
    def model(
        self,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ) -> ModelDF:
        model = ModelDF()
        model.agents.add([fix1_AgentSetPandas, fix2_AgentSetPolars])
        return model

    @pytest.fixture
    def grid_moore(self, model: ModelDF) -> GridPolars:
        space = GridPolars(model, dimensions=[3, 3], capacity=2)
        space.place_agents(agents=[0, 1], pos=[[0, 0], [1, 1]])
        space.set_cells(
            [[0, 0], [1, 1]], properties={"capacity": [1, 3], "property_0": "value_0"}
        )
        return space

    @pytest.fixture
    def grid_moore_torus(self, model: ModelDF) -> GridPolars:
        space = GridPolars(model, dimensions=[3, 3], capacity=2, torus=True)
        space.place_agents(agents=[0, 1], pos=[[0, 0], [1, 1]])
        space.set_cells(
            [[0, 0], [1, 1]], properties={"capacity": [1, 3], "property_0": "value_0"}
        )
        return space

    @pytest.fixture
    def grid_von_neumann(self, model: ModelDF) -> GridPolars:
        space = GridPolars(model, dimensions=[3, 3], neighborhood_type="von_neumann")
        space.place_agents(agents=[0, 1], pos=[[0, 0], [1, 1]])
        return space

    @pytest.fixture
    def grid_hexagonal(self, model: ModelDF) -> GridPolars:
        space = GridPolars(model, dimensions=[10, 10], neighborhood_type="hexagonal")
        space.place_agents(agents=[0, 1], pos=[[5, 4], [5, 5]])
        return space

    def test___init__(self, model: ModelDF):
        # Test with default parameters
        grid1 = GridPolars(model, dimensions=[3, 3])
        assert isinstance(grid1, GridPolars)
        assert isinstance(grid1.agents, pl.DataFrame)
        assert grid1.agents.is_empty()
        assert isinstance(grid1.cells, pl.DataFrame)
        assert grid1.cells.is_empty()
        assert isinstance(grid1.dimensions, list)
        assert len(grid1.dimensions) == 2
        assert isinstance(grid1.neighborhood_type, str)
        assert grid1.neighborhood_type == "moore"
        assert grid1.remaining_capacity == float("inf")
        assert grid1.model == model

        # Test with capacity = 10
        grid2 = GridPolars(model, dimensions=[3, 3], capacity=10)
        assert grid2.remaining_capacity == (10 * 3 * 3)

        # Test with torus = True
        grid3 = GridPolars(model, dimensions=[3, 3], torus=True)
        assert grid3.torus

        # Test with neighborhood_type = "von_neumann"
        grid4 = GridPolars(model, dimensions=[3, 3], neighborhood_type="von_neumann")
        assert grid4.neighborhood_type == "von_neumann"

        # Test with neighborhood_type = "moore"
        grid5 = GridPolars(model, dimensions=[3, 3], neighborhood_type="moore")
        assert grid5.neighborhood_type == "moore"

        # Test with neighborhood_type = "hexagonal"
        grid6 = GridPolars(model, dimensions=[3, 3], neighborhood_type="hexagonal")
        assert grid6.neighborhood_type == "hexagonal"

    def test_get_cells(self, grid_moore: GridPolars):
        # Test with None (all cells)
        result = grid_moore.get_cells()
        assert isinstance(result, pl.DataFrame)
        assert result.select(pl.col("dim_0")).to_series().to_list() == [0, 1]
        assert result.select(pl.col("dim_1")).to_series().to_list() == [0, 1]
        assert result.select(pl.col("capacity")).to_series().to_list() == [1, 3]
        assert result.select(pl.col("property_0")).to_series().to_list() == [
            "value_0",
            "value_0",
        ]

        # Test with GridCoordinate
        result = grid_moore.get_cells([0, 0])
        assert isinstance(result, pl.DataFrame)
        assert result.select(pl.col("dim_0")).to_series().to_list() == [0]
        assert result.select(pl.col("dim_1")).to_series().to_list() == [0]
        assert result.select(pl.col("capacity")).to_series().to_list() == [1]
        assert result.select(pl.col("property_0")).to_series().to_list() == ["value_0"]

        # Test with GridCoordinates
        result = grid_moore.get_cells([[0, 0], [1, 1]])
        assert isinstance(result, pl.DataFrame)
        assert result.select(pl.col("dim_0")).to_series().to_list() == [0, 1]
        assert result.select(pl.col("dim_1")).to_series().to_list() == [0, 1]
        assert result.select(pl.col("capacity")).to_series().to_list() == [1, 3]
        assert result.select(pl.col("property_0")).to_series().to_list() == [
            "value_0",
            "value_0",
        ]

    def test_get_directions(
        self,
        grid_moore: GridPolars,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ):
        # Test with GridCoordinate
        dir = grid_moore.get_directions(pos0=[1, 1], pos1=[2, 2])
        assert isinstance(dir, pl.DataFrame)
        assert dir.select(pl.col("dim_0")).to_series().to_list() == [1]
        assert dir.select(pl.col("dim_1")).to_series().to_list() == [1]

        # Test with GridCoordinates
        dir = grid_moore.get_directions(pos0=[[0, 0], [2, 2]], pos1=[[1, 2], [1, 1]])
        assert isinstance(dir, pl.DataFrame)
        assert dir.select(pl.col("dim_0")).to_series().to_list() == [1, -1]
        assert dir.select(pl.col("dim_1")).to_series().to_list() == [2, -1]

        # Test with missing agents (raises ValueError)
        with pytest.raises(ValueError):
            grid_moore.get_directions(
                agents0=fix1_AgentSetPandas, agents1=fix2_AgentSetPolars
            )

        # Test with IdsLike
        grid_moore.place_agents(fix2_AgentSetPolars, [[0, 1], [0, 2], [1, 0], [1, 2]])
        agents = grid_moore.agents.sort("agent_id")

        assert agents["agent_id"].to_list() == [0, 1, 4, 5, 6, 7]
        assert agents["dim_0"].to_list() == [0, 1, 0, 0, 1, 1]
        assert agents["dim_1"].to_list() == [0, 1, 1, 2, 0, 2]
        dir = grid_moore.get_directions(agents0=[0, 1], agents1=[4, 5])
        print(dir)
        assert isinstance(dir, pl.DataFrame)
        assert dir.select(pl.col("dim_0")).to_series().to_list() == [0, -1]
        assert dir.select(pl.col("dim_1")).to_series().to_list() == [1, 1]

        # Test with two AgentSetDFs
        grid_moore.place_agents([2, 3], [[1, 1], [2, 2]])
        dir = grid_moore.get_directions(
            agents0=fix1_AgentSetPandas, agents1=fix2_AgentSetPolars
        )
        assert isinstance(dir, pl.DataFrame)
        assert dir.select(pl.col("dim_0")).to_series().to_list() == [0, -1, 0, -1]
        assert dir.select(pl.col("dim_1")).to_series().to_list() == [1, 1, -1, 0]

        # Test with AgentsDF
        dir = grid_moore.get_directions(
            agents0=grid_moore.model.agents, agents1=grid_moore.model.agents
        )
        assert isinstance(dir, pl.DataFrame)
        assert grid_moore._df_all(dir == 0).all()

        # Test with normalize
        dir = grid_moore.get_directions(agents0=[0, 1], agents1=[4, 5], normalize=True)
        # Check if the vectors are normalized (length should be 1)
        assert np.allclose(
            np.sqrt(
                dir.select(pl.col("dim_0")).to_series().to_numpy() ** 2
                + dir.select(pl.col("dim_1")).to_series().to_numpy() ** 2
            ),
            1.0,
        )
        # Check specific normalized values
        assert np.allclose(
            dir.select(pl.col("dim_0")).to_series().to_list(), [0, -1 / np.sqrt(2)]
        )
        assert np.allclose(
            dir.select(pl.col("dim_1")).to_series().to_list(), [1, 1 / np.sqrt(2)]
        )

    def test_get_distances(
        self,
        grid_moore: GridPolars,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ):
        # Test with GridCoordinate
        dist = grid_moore.get_distances(pos0=[1, 1], pos1=[2, 2])
        assert isinstance(dist, pl.DataFrame)
        assert np.allclose(
            dist.select(pl.col("distance")).to_series().to_list(), [np.sqrt(2)]
        )

        # Test with GridCoordinates
        dist = grid_moore.get_distances(pos0=[[0, 0], [2, 2]], pos1=[[1, 2], [1, 1]])
        assert isinstance(dist, pl.DataFrame)
        assert np.allclose(
            dist.select(pl.col("distance")).to_series().to_list(),
            [np.sqrt(5), np.sqrt(2)],
        )

        # Test with missing agents (raises ValueError)
        with pytest.raises(ValueError):
            grid_moore.get_distances(
                agents0=fix1_AgentSetPandas, agents1=fix2_AgentSetPolars
            )

        # Test with IdsLike
        grid_moore.place_agents(fix2_AgentSetPolars, [[0, 1], [0, 2], [1, 0], [1, 2]])
        dist = grid_moore.get_distances(agents0=[0, 1], agents1=[4, 5])
        assert isinstance(dist, pl.DataFrame)
        assert np.allclose(
            dist.select(pl.col("distance")).to_series().to_list(), [1.0, np.sqrt(2)]
        )

        # Test with two AgentSetDFs
        grid_moore.place_agents([2, 3], [[1, 1], [2, 2]])
        dist = grid_moore.get_distances(
            agents0=fix1_AgentSetPandas, agents1=fix2_AgentSetPolars
        )
        assert isinstance(dist, pl.DataFrame)
        assert np.allclose(
            dist.select(pl.col("distance")).to_series().to_list(),
            [1.0, np.sqrt(2), 1.0, 1.0],
        )

        # Test with AgentsDF
        dist = grid_moore.get_distances(
            agents0=grid_moore.model.agents, agents1=grid_moore.model.agents
        )
        assert grid_moore._df_all(dist == 0).all()

    def test_get_neighborhood(
        self,
        grid_moore: GridPolars,
        grid_hexagonal: GridPolars,
        grid_von_neumann: GridPolars,
        grid_moore_torus: GridPolars,
    ):
        # Test with radius = int, pos=GridCoordinate
        neighborhood = grid_moore.get_neighborhood(radius=1, pos=[1, 1])
        assert isinstance(neighborhood, pl.DataFrame)
        assert set(neighborhood.columns) == {
            "dim_0",
            "dim_1",
            "radius",
            "dim_0_center",
            "dim_1_center",
        }
        assert neighborhood.shape == (8, 5)
        assert neighborhood.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            2,
        ]
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            2,
            0,
            1,
            2,
        ]
        assert neighborhood.select(pl.col("radius")).to_series().to_list() == [1] * 8
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list() == [1] * 8
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list() == [1] * 8
        )

        # Test with Sequence[int], pos=Sequence[GridCoordinate]
        neighborhood = grid_moore.get_neighborhood(radius=[1, 2], pos=[[1, 1], [2, 2]])
        neighborhood = neighborhood.sort(["dim_0_center", "dim_1_center", "radius"])
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (8 + 6, 5)
        assert (
            neighborhood.select(pl.col("radius")).to_series().to_list()
            == [1] * 11 + [2] * 3
        )
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list()
            == [1] * 8 + [2] * 6
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list()
            == [1] * 8 + [2] * 6
        )
        neighborhood = neighborhood.sort(["dim_0", "dim_1"])
        assert (
            neighborhood.select(pl.col("dim_0")).to_series().to_list()
            == [0] * 5 + [1] * 4 + [2] * 5
        )
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            0,
            1,
            2,
            2,
            0,
            1,
            2,
            2,
            0,
            0,
            1,
            1,
            2,
        ]

        grid_moore.place_agents([0, 1], [[1, 1], [2, 2]])

        # Test with agent=int, pos=GridCoordinate
        neighborhood = grid_moore.get_neighborhood(radius=1, agents=0)
        neighborhood = neighborhood.sort(["dim_0_center", "dim_1_center", "radius"])
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (8, 5)
        assert neighborhood.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            2,
        ]
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            2,
            0,
            1,
            2,
        ]
        assert neighborhood.select(pl.col("radius")).to_series().to_list() == [1] * 8
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list() == [1] * 8
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list() == [1] * 8
        )

        # Test with agent=Sequence[int], pos=Sequence[GridCoordinate]
        neighborhood = grid_moore.get_neighborhood(radius=[1, 2], agents=[0, 1])
        neighborhood = neighborhood.sort(["dim_0_center", "dim_1_center", "radius"])
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (8 + 6, 5)
        assert (
            neighborhood.select(pl.col("radius")).to_series().to_list()
            == [1] * 11 + [2] * 3
        )
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list()
            == [1] * 8 + [2] * 6
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list()
            == [1] * 8 + [2] * 6
        )
        neighborhood = neighborhood.sort(["dim_0", "dim_1"])
        assert (
            neighborhood.select(pl.col("dim_0")).to_series().to_list()
            == [0] * 5 + [1] * 4 + [2] * 5
        )
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            0,
            1,
            2,
            2,
            0,
            1,
            2,
            2,
            0,
            0,
            1,
            1,
            2,
        ]

        # Test with include_center
        neighborhood = grid_moore.get_neighborhood(
            radius=1, pos=[1, 1], include_center=True
        )
        neighborhood = neighborhood.sort(["dim_0_center", "dim_1_center", "radius"])
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (9, 5)
        assert neighborhood.select(pl.col("dim_0")).to_series().to_list() == [
            1,
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            2,
        ]
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            1,
            0,
            1,
            2,
            0,
            2,
            0,
            1,
            2,
        ]
        assert (
            neighborhood.select(pl.col("radius")).to_series().to_list() == [0] + [1] * 8
        )
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list() == [1] * 9
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list() == [1] * 9
        )

        # Test with torus
        neighborhood = grid_moore_torus.get_neighborhood(radius=1, pos=[0, 0])
        neighborhood = neighborhood.sort(["dim_0_center", "dim_1_center", "radius"])
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (8, 5)
        assert neighborhood.select(pl.col("dim_0")).to_series().to_list() == [
            2,
            2,
            2,
            0,
            0,
            1,
            1,
            1,
        ]
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            2,
            0,
            1,
            2,
            1,
            2,
            0,
            1,
        ]
        assert neighborhood.select(pl.col("radius")).to_series().to_list() == [1] * 8
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list() == [0] * 8
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list() == [0] * 8
        )

        # Test with radius and pos of different length
        with pytest.raises(ValueError):
            neighborhood = grid_moore.get_neighborhood(radius=[1, 2], pos=[1, 1])

        # Test with von_neumann neighborhood
        neighborhood = grid_von_neumann.get_neighborhood(radius=1, pos=[1, 1])
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (4, 5)
        assert neighborhood.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            1,
            1,
            2,
        ]
        assert neighborhood.select(pl.col("dim_1")).to_series().to_list() == [
            1,
            0,
            2,
            1,
        ]
        assert neighborhood.select(pl.col("radius")).to_series().to_list() == [1] * 4
        assert (
            neighborhood.select(pl.col("dim_0_center")).to_series().to_list() == [1] * 4
        )
        assert (
            neighborhood.select(pl.col("dim_1_center")).to_series().to_list() == [1] * 4
        )

        # Test with hexagonal neighborhood (odd cell [2,1] and even cell [2,2])
        neighborhood = grid_hexagonal.get_neighborhood(
            radius=[2, 3], pos=[[5, 4], [5, 5]]
        )
        assert isinstance(neighborhood, pl.DataFrame)
        assert neighborhood.shape == (
            6 * 2 + 12 * 2 + 18,
            5,
        )  # 6 neighbors for radius 1, 12 for radius 2, 18 for radius 3

        # Sort the neighborhood for consistent ordering
        neighborhood = neighborhood.sort(
            ["dim_0_center", "dim_1_center", "radius", "dim_0", "dim_1"]
        )

        # Expected neighbors for [5,4] and [5,5]
        expected_neighbors = [
            # Neighbors of [5,4]
            # radius 1
            (4, 4),
            (4, 5),
            (5, 3),
            (5, 5),
            (6, 3),
            (6, 4),
            # radius 2
            (3, 4),
            (3, 6),
            (4, 2),
            (4, 5),
            (4, 6),
            (5, 2),
            (5, 5),
            (5, 6),
            (6, 3),
            (7, 2),
            (7, 3),
            (7, 4),
            # Neighbors of [5,5]
            # radius 1
            (4, 5),
            (4, 6),
            (5, 4),
            (5, 6),
            (6, 4),
            (6, 5),
            # radius 2
            (3, 5),
            (3, 7),
            (4, 3),
            (4, 6),
            (4, 7),
            (5, 3),
            (5, 6),
            (5, 7),
            (6, 4),
            (7, 3),
            (7, 4),
            (7, 5),
            # radius 3
            (2, 5),
            (2, 8),
            (3, 2),
            (3, 6),
            (3, 8),
            (4, 2),
            (4, 7),
            (4, 8),
            (5, 2),
            (5, 6),
            (5, 7),
            (5, 8),
            (6, 3),
            (7, 4),
            (8, 2),
            (8, 3),
            (8, 4),
            (8, 5),
        ]

        assert (
            list(
                zip(
                    neighborhood.select(pl.col("dim_0")).to_series().to_list(),
                    neighborhood.select(pl.col("dim_1")).to_series().to_list(),
                )
            )
            == expected_neighbors
        )

    def test_get_neighbors(
        self,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
        grid_moore: GridPolars,
        grid_hexagonal: GridPolars,
        grid_von_neumann: GridPolars,
        grid_moore_torus: GridPolars,
    ):
        # Place agents in the grid
        grid_moore.move_agents(
            [0, 1, 2, 3, 4, 5, 6, 7],
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 2], [2, 0], [2, 1], [2, 2]],
        )

        # Test with radius = int, pos=GridCoordinate
        neighbors = grid_moore.get_neighbors(radius=1, pos=[1, 1])
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.columns == ["agent_id", "dim_0", "dim_1"]
        assert neighbors.shape == (8, 3)
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            2,
        ]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            2,
            0,
            1,
            2,
        ]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        }

        # Test with Sequence[int], pos=Sequence[GridCoordinate]
        neighbors = grid_moore.get_neighbors(radius=[1, 2], pos=[[1, 1], [2, 2]])
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (8, 3)
        neighbors = neighbors.sort(["dim_0", "dim_1"])
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            2,
        ]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            2,
            0,
            1,
            2,
        ]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        }

        # Test with agent=int
        neighbors = grid_moore.get_neighbors(radius=1, agents=0)
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (2, 3)
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [0, 1]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [1, 0]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {1, 3}

        # Test with agent=Sequence[int]
        neighbors = grid_moore.get_neighbors(radius=[1, 2], agents=[0, 7])
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (7, 3)
        neighbors = neighbors.sort(["dim_0", "dim_1"])
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            0,
            0,
            1,
            1,
            2,
            2,
        ]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            2,
            0,
            1,
        ]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {
            0,
            1,
            2,
            3,
            4,
            5,
            6,
        }

        # Test with include_center
        neighbors = grid_moore.get_neighbors(radius=1, pos=[1, 1], include_center=True)
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (8, 3)  # No agent at [1, 1], so still 8 neighbors
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            0,
            0,
            1,
            1,
            2,
            2,
            2,
        ]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            2,
            0,
            1,
            2,
        ]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        }

        # Test with torus
        grid_moore_torus.move_agents(
            [0, 1, 2, 3, 4, 5, 6, 7],
            [[2, 2], [2, 0], [2, 1], [0, 2], [0, 1], [1, 2], [1, 0], [1, 1]],
        )
        neighbors = grid_moore_torus.get_neighbors(radius=1, pos=[0, 0])
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (8, 3)
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [
            2,
            2,
            2,
            0,
            0,
            1,
            1,
            1,
        ]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [
            2,
            0,
            1,
            2,
            1,
            2,
            0,
            1,
        ]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        }

        # Test with radius and pos of different length
        with pytest.raises(ValueError):
            neighbors = grid_moore.get_neighbors(radius=[1, 2], pos=[1, 1])

        # Test with von_neumann neighborhood
        grid_von_neumann.move_agents([0, 1, 2, 3], [[0, 1], [1, 0], [1, 2], [2, 1]])
        neighbors = grid_von_neumann.get_neighbors(radius=1, pos=[1, 1])
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (4, 3)
        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [0, 1, 1, 2]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [1, 0, 2, 1]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == {
            0,
            1,
            2,
            3,
        }

        # Test with hexagonal neighborhood (odd cell [5,4] and even cell [5,5])
        grid_hexagonal.move_agents(
            range(8), [[4, 4], [4, 5], [5, 3], [5, 5], [6, 3], [6, 4], [5, 4], [5, 6]]
        )
        neighbors = grid_hexagonal.get_neighbors(radius=[2, 3], pos=[[5, 4], [5, 5]])
        assert isinstance(neighbors, pl.DataFrame)
        assert neighbors.shape == (8, 3)  # All agents are within the neighborhood

        # Sort the neighbors for consistent ordering
        neighbors = neighbors.sort(["dim_0", "dim_1"])

        assert neighbors.select(pl.col("dim_0")).to_series().to_list() == [
            4,
            4,
            5,
            5,
            5,
            5,
            6,
            6,
        ]
        assert neighbors.select(pl.col("dim_1")).to_series().to_list() == [
            4,
            5,
            3,
            4,
            5,
            6,
            3,
            4,
        ]
        assert set(neighbors.select(pl.col("agent_id")).to_series().to_list()) == set(
            range(8)
        )

    def test_is_available(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        result = grid_moore.is_available([0, 0])
        assert isinstance(result, pl.DataFrame)
        assert result.select(pl.col("available")).to_series().to_list() == [False]
        result = grid_moore.is_available([1, 1])
        assert result.select(pl.col("available")).to_series().to_list() == [True]

        # Test with GridCoordinates
        result = grid_moore.is_available([[0, 0], [1, 1]])
        assert result.select(pl.col("available")).to_series().to_list() == [False, True]

    def test_is_empty(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        result = grid_moore.is_empty([0, 0])
        assert isinstance(result, pl.DataFrame)
        assert result.select(pl.col("empty")).to_series().to_list() == [False]
        result = grid_moore.is_empty([1, 1])
        assert result.select(pl.col("empty")).to_series().to_list() == [False]

        # Test with GridCoordinates
        result = grid_moore.is_empty([[0, 0], [1, 1]])
        assert result.select(pl.col("empty")).to_series().to_list() == [False, False]

    def test_is_full(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        result = grid_moore.is_full([0, 0])
        assert isinstance(result, pl.DataFrame)
        assert result.select(pl.col("full")).to_series().to_list() == [True]
        result = grid_moore.is_full([1, 1])
        assert result.select(pl.col("full")).to_series().to_list() == [False]

        # Test with GridCoordinates
        result = grid_moore.is_full([[0, 0], [1, 1]])
        assert result.select(pl.col("full")).to_series().to_list() == [True, False]

    def test_move_agents(
        self,
        grid_moore: GridPolars,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ):
        # Test with IdsLike
        space = grid_moore.move_agents(agents=1, pos=[1, 1], inplace=False)
        assert space.remaining_capacity == (2 * 3 * 3 - 2)
        assert len(space.agents) == 2
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [0, 1]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [0, 1]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [0, 1]

        # Test with AgentSetDF
        with pytest.warns(RuntimeWarning):
            space = grid_moore.move_agents(
                agents=fix2_AgentSetPolars,
                pos=[[0, 0], [1, 0], [2, 0], [0, 1]],
                inplace=False,
            )
            assert space.remaining_capacity == (2 * 3 * 3 - 6)
            assert len(space.agents) == 6
            agents = space.agents.sort("agent_id")
            assert agents.select(pl.col("agent_id")).to_series().to_list() == [
                0,
                1,
                4,
                5,
                6,
                7,
            ]
            assert agents.select(pl.col("dim_0")).to_series().to_list() == [
                0,
                1,
                0,
                1,
                2,
                0,
            ]
            assert agents.select(pl.col("dim_1")).to_series().to_list() == [
                0,
                1,
                0,
                0,
                0,
                1,
            ]

        # Test with Collection[AgentSetDF]
        with pytest.warns(RuntimeWarning):
            space = grid_moore.move_agents(
                agents=[fix1_AgentSetPandas, fix2_AgentSetPolars],
                pos=[[0, 2], [1, 2], [2, 2], [0, 1], [1, 1], [2, 1], [0, 0], [1, 0]],
                inplace=False,
            )
        assert space.remaining_capacity == (2 * 3 * 3 - 8)
        assert len(space.agents) == 8
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        ]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            1,
            2,
            0,
            1,
        ]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [
            2,
            2,
            2,
            1,
            1,
            1,
            0,
            0,
        ]

        # Raises ValueError if len(agents) != len(pos)
        with pytest.raises(ValueError):
            space = grid_moore.move_agents(
                agents=[0, 1], pos=[[0, 0], [1, 1], [2, 2]], inplace=False
            )

        # Test with AgentsDF, pos=DataFrame
        pos = pl.DataFrame(
            {
                "dim_0": [0, 1, 2, 0, 1, 2, 0, 1],
                "dim_1": [2, 2, 2, 1, 1, 1, 0, 0],
            }
        )

        with pytest.warns(RuntimeWarning):
            space = grid_moore.move_agents(
                agents=grid_moore.model.agents,
                pos=pos,
                inplace=False,
            )
        assert space.remaining_capacity == (2 * 3 * 3 - 8)
        assert len(space.agents) == 8
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        ]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            1,
            2,
            0,
            1,
        ]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [
            2,
            2,
            2,
            1,
            1,
            1,
            0,
            0,
        ]

        # Test with agents=int, pos=DataFrame
        pos = pl.DataFrame({"dim_0": [0], "dim_1": [2]})
        space = grid_moore.move_agents(agents=1, pos=pos, inplace=False)
        assert space.remaining_capacity == (2 * 3 * 3 - 2)
        assert len(space.agents) == 2
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [0, 1]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [0, 0]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [0, 2]

    def test_move_to_available(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        last = None
        different = False
        for _ in range(10):
            available_cells = grid_moore.available_cells
            space = grid_moore.move_to_available(0, inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in available_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with GridCoordinates
        last = None
        different = False
        for _ in range(10):
            available_cells = grid_moore.available_cells
            space = grid_moore.move_to_available([0, 1], inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in available_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in available_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with AgentSetDF
        last = None
        different = False
        for _ in range(10):
            available_cells = grid_moore.available_cells
            space = grid_moore.move_to_available(grid_moore.model.agents, inplace=False)
            if last is not None and not different:
                if (space.agents.select(pl.col("dim_0")).to_numpy() != last).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in available_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in available_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0")).to_numpy()
        assert different

    def test_move_to_empty(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        last = None
        different = False
        for _ in range(10):
            empty_cells = grid_moore.empty_cells
            space = grid_moore.move_to_empty(0, inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.filter(pl.col("agent_id") == 0).row(0)[1:]
                in empty_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with GridCoordinates
        last = None
        different = False
        for _ in range(10):
            empty_cells = grid_moore.empty_cells
            space = grid_moore.move_to_empty([0, 1], inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in empty_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in empty_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with AgentSetDF
        last = None
        different = False
        for _ in range(10):
            empty_cells = grid_moore.empty_cells
            space = grid_moore.move_to_empty(grid_moore.model.agents, inplace=False)
            if last is not None and not different:
                if (space.agents.select(pl.col("dim_0")).to_numpy() != last).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in empty_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in empty_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0")).to_numpy()
        assert different

    def test_out_of_bounds(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        out_of_bounds = grid_moore.out_of_bounds([11, 11])
        assert isinstance(out_of_bounds, pl.DataFrame)
        assert out_of_bounds.shape == (1, 3)
        assert out_of_bounds.columns == ["dim_0", "dim_1", "out_of_bounds"]
        assert out_of_bounds.row(0) == (11, 11, True)

        # Test with GridCoordinates
        out_of_bounds = grid_moore.out_of_bounds([[0, 0], [11, 11]])
        assert isinstance(out_of_bounds, pl.DataFrame)
        assert out_of_bounds.shape == (2, 3)
        assert out_of_bounds.columns == ["dim_0", "dim_1", "out_of_bounds"]
        assert out_of_bounds.row(0) == (0, 0, False)
        assert out_of_bounds.row(1) == (11, 11, True)

    def test_place_agents(
        self,
        grid_moore: GridPolars,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ):
        # Test with IdsLike
        with pytest.warns(RuntimeWarning):
            space = grid_moore.place_agents(
                agents=[1, 2], pos=[[1, 1], [2, 2]], inplace=False
            )
        assert space.remaining_capacity == (2 * 3 * 3 - 3)
        assert len(space.agents) == 3
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
            2,
        ]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [0, 1, 2]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [0, 1, 2]

        # Test with agents not in the model
        with pytest.raises(ValueError):
            space = grid_moore.place_agents(
                agents=[10, 11],
                pos=[[0, 0], [1, 0]],
                inplace=False,
            )

        # Test with AgentSetDF
        space = grid_moore.place_agents(
            agents=fix2_AgentSetPolars,
            pos=[[0, 0], [1, 0], [2, 0], [0, 1]],
            inplace=False,
        )
        assert space.remaining_capacity == (2 * 3 * 3 - 6)
        assert len(space.agents) == 6
        agents = space.agents.sort("agent_id")
        assert agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
            4,
            5,
            6,
            7,
        ]
        assert agents.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            1,
            0,
            1,
            2,
            0,
        ]
        assert agents.select(pl.col("dim_1")).to_series().to_list() == [
            0,
            1,
            0,
            0,
            0,
            1,
        ]

        # Test with Collection[AgentSetDF]
        with pytest.warns(RuntimeWarning):
            space = grid_moore.place_agents(
                agents=[fix1_AgentSetPandas, fix2_AgentSetPolars],
                pos=[[0, 2], [1, 2], [2, 2], [0, 1], [1, 1], [2, 1], [0, 0], [1, 0]],
                inplace=False,
            )
        assert space.remaining_capacity == (2 * 3 * 3 - 8)
        assert len(space.agents) == 8
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        ]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            1,
            2,
            0,
            1,
        ]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [
            2,
            2,
            2,
            1,
            1,
            1,
            0,
            0,
        ]

        # Test with AgentsDF, pos=DataFrame
        pos = pl.DataFrame(
            {
                "dim_0": [0, 1, 2, 0, 1, 2, 0, 1],
                "dim_1": [2, 2, 2, 1, 1, 1, 0, 0],
            }
        )
        with pytest.warns(RuntimeWarning):
            space = grid_moore.place_agents(
                agents=grid_moore.model.agents,
                pos=pos,
                inplace=False,
            )
        assert space.remaining_capacity == (2 * 3 * 3 - 8)
        assert len(space.agents) == 8
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
        ]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [
            0,
            1,
            2,
            0,
            1,
            2,
            0,
            1,
        ]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [
            2,
            2,
            2,
            1,
            1,
            1,
            0,
            0,
        ]

        # Test with agents=int, pos=DataFrame
        pos = pl.DataFrame({"dim_0": [0], "dim_1": [2]})
        with pytest.warns(RuntimeWarning):
            space = grid_moore.place_agents(agents=1, pos=pos, inplace=False)
        assert space.remaining_capacity == (2 * 3 * 3 - 2)
        assert len(space.agents) == 2
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [0, 1]
        assert space.agents.select(pl.col("dim_0")).to_series().to_list() == [0, 0]
        assert space.agents.select(pl.col("dim_1")).to_series().to_list() == [0, 2]

    def test_place_to_available(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        last = None
        different = False
        for _ in range(10):
            available_cells = grid_moore.available_cells
            space = grid_moore.place_to_available(0, inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in available_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with GridCoordinates
        last = None
        different = False
        for _ in range(10):
            available_cells = grid_moore.available_cells
            space = grid_moore.place_to_available([0, 1], inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in available_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in available_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with AgentSetDF
        last = None
        different = False
        for _ in range(10):
            available_cells = grid_moore.available_cells
            space = grid_moore.place_to_available(
                grid_moore.model.agents, inplace=False
            )
            if last is not None and not different:
                if (space.agents.select(pl.col("dim_0")).to_numpy() != last).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in available_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in available_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0")).to_numpy()
        assert different

    def test_place_to_empty(self, grid_moore: GridPolars):
        # Test with GridCoordinate
        last = None
        different = False
        for _ in range(10):
            empty_cells = grid_moore.empty_cells
            space = grid_moore.place_to_empty(0, inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.filter(pl.col("agent_id") == 0).row(0)[1:]
                in empty_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with GridCoordinates
        last = None
        different = False
        for _ in range(10):
            empty_cells = grid_moore.empty_cells
            space = grid_moore.place_to_empty([0, 1], inplace=False)
            if last is not None and not different:
                if (
                    space.agents.select(pl.col("dim_0", "dim_1")).to_numpy() != last
                ).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in empty_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in empty_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0", "dim_1")).to_numpy()
        assert different

        # Test with AgentSetDF
        last = None
        different = False
        for _ in range(10):
            empty_cells = grid_moore.empty_cells
            space = grid_moore.place_to_empty(grid_moore.model.agents, inplace=False)
            if last is not None and not different:
                if (space.agents.select(pl.col("dim_0")).to_numpy() != last).any():
                    different = True
            assert (
                space.agents.select(pl.col("dim_0", "dim_1")).row(0)
                in empty_cells.rows()
            ) and (
                space.agents.select(pl.col("dim_0", "dim_1")).row(1)
                in empty_cells.rows()
            )
            last = space.agents.select(pl.col("dim_0")).to_numpy()
        assert different

    def test_random_agents(self, grid_moore: GridPolars):
        different = False
        agents0 = grid_moore.random_agents(1)
        for _ in range(100):
            agents1 = grid_moore.random_agents(1)
            if (agents0.to_numpy() != agents1.to_numpy()).all():
                different = True
                break
        assert different

    def test_random_pos(self, grid_moore: GridPolars):
        different = False
        last = None
        for _ in range(10):
            random_pos = grid_moore.random_pos(5)
            assert isinstance(random_pos, pl.DataFrame)
            assert len(random_pos) == 5
            assert random_pos.columns == ["dim_0", "dim_1"]
            assert (
                not grid_moore.out_of_bounds(random_pos)
                .select(pl.col("out_of_bounds"))
                .to_series()
                .any()
            )
            if last is not None and not different:
                if (last.to_numpy() != random_pos.to_numpy()).any():
                    different = True
                break
            last = random_pos
        assert different

    def test_remove_agents(
        self,
        grid_moore: GridPolars,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ):
        grid_moore.move_agents(
            [0, 1, 2, 3, 4, 5, 6, 7],
            [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]],
        )
        capacity = grid_moore.remaining_capacity
        # Test with IdsLike
        space = grid_moore.remove_agents([1, 2], inplace=False)
        assert space.agents.shape == (6, 3)
        assert space.remaining_capacity == capacity + 2
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            3,
            4,
            5,
            6,
            7,
        ]
        assert [
            x for id in space.model.agents.index.values() for x in id.to_list()
        ] == [x for x in range(8)]

        # Test with AgentSetDF
        space = grid_moore.remove_agents(fix1_AgentSetPandas, inplace=False)
        assert space.agents.shape == (4, 3)
        assert space.remaining_capacity == capacity + 4
        assert space.agents.select(pl.col("agent_id")).to_series().to_list() == [
            4,
            5,
            6,
            7,
        ]
        assert [
            x for id in space.model.agents.index.values() for x in id.to_list()
        ] == [x for x in range(8)]

        # Test with Collection[AgentSetDF]
        space = grid_moore.remove_agents(
            [fix1_AgentSetPandas, fix2_AgentSetPolars], inplace=False
        )
        assert [
            x for id in space.model.agents.index.values() for x in id.to_list()
        ] == [x for x in range(8)]
        assert space.agents.is_empty()
        assert space.remaining_capacity == capacity + 8
        # Test with AgentsDF
        space = grid_moore.remove_agents(grid_moore.model.agents, inplace=False)
        assert space.remaining_capacity == capacity + 8
        assert space.agents.is_empty()
        assert [
            x for id in space.model.agents.index.values() for x in id.to_list()
        ] == [x for x in range(8)]

    def test_sample_cells(self, grid_moore: GridPolars):
        # Test with default parameters
        replacement = False
        same = True
        last = None
        for _ in range(10):
            result = grid_moore.sample_cells(10)
            assert len(result) == 10
            assert isinstance(result, pl.DataFrame)
            assert result.columns == ["dim_0", "dim_1"]
            counts = result.group_by("dim_0", "dim_1").agg(pl.count())
            assert (counts.select(pl.col("count")) <= 2).to_series().all()
            if (
                not replacement
                and (counts.select(pl.col("count")) > 1).to_series().any()
            ):
                replacement = True
            if same and last is not None:
                same = (result.to_numpy() == last).all()
            if not same and replacement:
                break
            last = result.to_numpy()
        assert replacement and not same

        # Test with too many samples
        with pytest.raises(AssertionError):
            grid_moore.sample_cells(100)

        # Test with 'empty' cell_type
        result = grid_moore.sample_cells(14, cell_type="empty")
        assert len(result) == 14
        assert isinstance(result, pl.DataFrame)
        assert result.columns == ["dim_0", "dim_1"]
        counts = result.group_by("dim_0", "dim_1").agg(pl.count())

        ## (0, 1) and (1, 1) are not in the result
        assert not (
            result.select(pl.col("dim_0") == 0).to_series()
            & (result.select(pl.col("dim_1") == 0)).to_series()
        ).any(), "Found (0, 1) in the result"
        assert not (
            result.select(pl.col("dim_0") == 1).to_series()
            & (result.select(pl.col("dim_1") == 1)).to_series()
        ).any(), "Found (1, 1) in the result"

        # 14 should be the max number of empty cells
        with pytest.raises(AssertionError):
            grid_moore.sample_cells(15, cell_type="empty")

        # Test with 'available' cell_type
        result = grid_moore.sample_cells(16, cell_type="available")
        assert len(result) == 16
        assert isinstance(result, pl.DataFrame)
        assert result.columns == ["dim_0", "dim_1"]
        counts = result.group_by("dim_0", "dim_1").agg(pl.count())

        # 16 should be the max number of available cells
        with pytest.raises(AssertionError):
            grid_moore.sample_cells(17, cell_type="available")

        # Test with 'full' cell_type and no replacement
        grid_moore.set_cells([[0, 0], [1, 1]], properties={"capacity": 1})
        result = grid_moore.sample_cells(2, cell_type="full", with_replacement=False)
        assert len(result) == 2
        assert isinstance(result, pl.DataFrame)
        assert result.columns == ["dim_0", "dim_1"]
        assert (
            (
                result.select(pl.col("dim_0") == 0).to_series()
                & (result.select(pl.col("dim_1") == 0)).to_series()
            )
            | (
                result.select(pl.col("dim_0") == 1).to_series()
                & (result.select(pl.col("dim_1") == 1)).to_series()
            )
        ).all()
        # 2 should be the max number of full cells
        with pytest.raises(AssertionError):
            grid_moore.sample_cells(3, cell_type="full", with_replacement=False)

    def test_set_cells(self, model: ModelDF):
        # Initialize GridPolars
        grid_moore = GridPolars(model, dimensions=[3, 3], capacity=2)

        # Test with GridCoordinate
        grid_moore.set_cells(
            [0, 0], properties={"capacity": 1, "property_0": "value_0"}
        )
        assert grid_moore.remaining_capacity == (2 * 3 * 3 - 1)
        cell_df = grid_moore.get_cells([0, 0])
        assert cell_df["capacity"][0] == 1
        assert cell_df["property_0"][0] == "value_0"

        # Test with GridCoordinates
        grid_moore.set_cells(
            [[1, 1], [2, 2]], properties={"capacity": 3, "property_1": "value_1"}
        )
        assert grid_moore.remaining_capacity == (2 * 3 * 3 - 1 + 2)
        cell_df = grid_moore.get_cells([[1, 1], [2, 2]])
        assert cell_df["capacity"][0] == 3
        assert cell_df["property_1"][0] == "value_1"
        assert cell_df["capacity"][1] == 3
        assert cell_df["property_1"][1] == "value_1"

        cell_df = grid_moore.get_cells([0, 0])
        assert cell_df["capacity"][0] == 1
        assert cell_df["property_0"][0] == "value_0"

        # Test with DataFrame
        df = pl.DataFrame(
            {"dim_0": [0, 1, 2], "dim_1": [0, 1, 2], "capacity": [2, 2, 2]}
        )
        grid_moore.set_cells(df)
        assert grid_moore.remaining_capacity == (2 * 3 * 3)

        cells_df = grid_moore.get_cells([[0, 0], [1, 1], [2, 2]])
        assert cells_df["capacity"][0] == 2
        assert cells_df["capacity"][1] == 2
        assert cells_df["capacity"][2] == 2
        assert cells_df["property_0"][0] == "value_0"
        assert cells_df["property_1"][1] == "value_1"
        assert cells_df["property_1"][2] == "value_1"

        # Add 2 agents to a cell, then set the cell capacity to 1
        grid_moore.place_agents([1, 2], [[0, 0], [0, 0]])
        with pytest.raises(AssertionError):
            grid_moore.set_cells([0, 0], properties={"capacity": 1})

    def test_swap_agents(
        self,
        grid_moore: GridPolars,
        fix1_AgentSetPandas: ExampleAgentSetPandas,
        fix2_AgentSetPolars: ExampleAgentSetPolars,
    ):
        grid_moore.move_agents(
            [0, 1, 2, 3, 4, 5, 6, 7],
            [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]],
        )
        # Test with IdsLike
        space = grid_moore.swap_agents([0, 1], [2, 3], inplace=False)
        assert (
            space.agents.filter(pl.col("agent_id") == 0).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 2).row(0)[1:]
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 1).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 3).row(0)[1:]
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 2).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 0).row(0)[1:]
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 3).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 1).row(0)[1:]
        )
        # Test with AgentSetDFs
        space = grid_moore.swap_agents(
            fix1_AgentSetPandas, fix2_AgentSetPolars, inplace=False
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 0).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 4).row(0)[1:]
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 1).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 5).row(0)[1:]
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 2).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 6).row(0)[1:]
        )
        assert (
            space.agents.filter(pl.col("agent_id") == 3).row(0)[1:]
            == grid_moore.agents.filter(pl.col("agent_id") == 7).row(0)[1:]
        )

    def test_torus_adj(self, grid_moore: GridPolars, grid_moore_torus: GridPolars):
        # Test with non-toroidal grid
        with pytest.raises(ValueError):
            grid_moore.torus_adj([10, 10])

        # Test with toroidal grid (GridCoordinate)
        adj_df = grid_moore_torus.torus_adj([10, 8])
        assert isinstance(adj_df, pl.DataFrame)
        assert adj_df.shape == (1, 2)
        assert adj_df.columns == ["dim_0", "dim_1"]
        assert adj_df.row(0) == (1, 2)

        # Test with toroidal grid (GridCoordinates)
        adj_df = grid_moore_torus.torus_adj([[10, 8], [15, 11]])
        assert isinstance(adj_df, pl.DataFrame)
        assert adj_df.shape == (2, 2)
        assert adj_df.columns == ["dim_0", "dim_1"]
        assert adj_df.row(0) == (1, 2)
        assert adj_df.row(1) == (0, 2)

    def test___getitem__(self, grid_moore: GridPolars):
        # Test out of bounds
        with pytest.raises(ValueError):
            grid_moore[[5, 5]]

        # Test with GridCoordinate
        df = grid_moore[[0, 0]]
        assert isinstance(df, pl.DataFrame)
        assert df.columns == ["dim_0", "dim_1", "capacity", "property_0", "agent_id"]
        assert df.row(0) == (0, 0, 1, "value_0", 0)

        # Test with GridCoordinates
        df = grid_moore[[[0, 0], [1, 1]]]
        assert isinstance(df, pl.DataFrame)
        assert df.columns == ["dim_0", "dim_1", "capacity", "property_0", "agent_id"]
        assert df.row(0) == (0, 0, 1, "value_0", 0)
        assert df.row(1) == (1, 1, 3, "value_0", 1)

    def test___setitem__(self, grid_moore: GridPolars):
        # Test with out-of-bounds
        with pytest.raises(ValueError):
            grid_moore[[5, 5]] = {"capacity": 10}

        # Test with GridCoordinate
        grid_moore[[0, 0]] = {"capacity": 10}
        assert grid_moore.get_cells([[0, 0]])["capacity"][0] == 10
        # Test with GridCoordinates
        grid_moore[[[0, 0], [1, 1]]] = {"capacity": 20}
        assert grid_moore.get_cells([[0, 0], [1, 1]]).select(
            pl.col("capacity")
        ).to_series().to_list() == [20, 20]

    # Property tests
    def test_agents(self, grid_moore: GridPolars):
        assert isinstance(grid_moore.agents, pl.DataFrame)
        assert grid_moore.agents.select(pl.col("agent_id")).to_series().to_list() == [
            0,
            1,
        ]
        assert grid_moore.agents.columns == ["agent_id", "dim_0", "dim_1"]
        assert grid_moore.agents.select(pl.col("dim_0")).to_series().to_list() == [0, 1]
        assert grid_moore.agents.select(pl.col("dim_1")).to_series().to_list() == [0, 1]

    def test_available_cells(self, grid_moore: GridPolars):
        result = grid_moore.available_cells
        assert len(result) == 8
        assert isinstance(result, pl.DataFrame)
        assert result.columns == ["dim_0", "dim_1"]

    def test_cells(self, grid_moore: GridPolars):
        result = grid_moore.cells
        assert isinstance(result, pl.DataFrame)
        assert result.columns == [
            "dim_0",
            "dim_1",
            "capacity",
            "property_0",
            "agent_id",
        ]
        assert result.select(pl.col("dim_0")).to_series().to_list() == [0, 1]
        assert result.select(pl.col("dim_1")).to_series().to_list() == [0, 1]
        assert result.select(pl.col("capacity")).to_series().to_list() == [1, 3]
        assert result.select(pl.col("property_0")).to_series().to_list() == [
            "value_0",
            "value_0",
        ]
        assert result.select(pl.col("agent_id")).to_series().to_list() == [0, 1]

    def test_dimensions(self, grid_moore: GridPolars):
        assert isinstance(grid_moore.dimensions, list)
        assert len(grid_moore.dimensions) == 2

    def test_empty_cells(self, grid_moore: GridPolars):
        result = grid_moore.empty_cells
        assert len(result) == 7
        assert isinstance(result, pl.DataFrame)
        assert result.columns == ["dim_0", "dim_1"]

    def test_full_cells(self, grid_moore: GridPolars):
        grid_moore.set_cells([[0, 0], [1, 1]], {"capacity": 1})
        result = grid_moore.full_cells
        assert len(result) == 2
        assert isinstance(result, pl.DataFrame)
        assert result.columns == ["dim_0", "dim_1"]
        assert (
            (
                result.select(pl.col("dim_0") == 0).to_series()
                & (result.select(pl.col("dim_1") == 0)).to_series()
            )
            | (
                result.select(pl.col("dim_0") == 1).to_series()
                & (result.select(pl.col("dim_1") == 1)).to_series()
            )
        ).all()

    def test_model(self, grid_moore: GridPolars, model: ModelDF):
        assert grid_moore.model == model

    def test_neighborhood_type(
        self,
        grid_moore: GridPolars,
        grid_von_neumann: GridPolars,
        grid_hexagonal: GridPolars,
    ):
        assert grid_moore.neighborhood_type == "moore"
        assert grid_von_neumann.neighborhood_type == "von_neumann"
        assert grid_hexagonal.neighborhood_type == "hexagonal"

    def test_random(self, grid_moore: GridPolars):
        assert grid_moore.random == grid_moore.model.random

    def test_remaining_capacity(self, grid_moore: GridPolars):
        assert grid_moore.remaining_capacity == (3 * 3 * 2 - 2)

    def test_torus(self, model: ModelDF, grid_moore: GridPolars):
        assert not grid_moore.torus

        grid_2 = GridPolars(model, [3, 3], torus=True)
        assert grid_2.torus
